﻿Imports System.IO
Imports System.Text
Imports System.Xml

Module MMPrj1

    Public Class MyUnits
        Public id As Integer
        Public Name As String
        Public ruName As String
        Public b As Double
        Public c As Double
        Public proj As String
        Public EPSG As Integer
        Public target_EPSG As Integer

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Unit")
            '
            res.SetAttribute("id", id)
            res.SetAttribute("Name", Name)
            res.SetAttribute("ruName", ruName)
            res.SetAttribute("b", b)
            res.SetAttribute("c", c)
            res.SetAttribute("proj", proj)
            res.SetAttribute("EPSG", EPSG)
            res.SetAttribute("target_EPSG", target_EPSG)
            '
            Return res
        End Function

        Public Function toProj() As String
            Return "+units=" & proj
        End Function

    End Class

    Public Class MyAffineParam
        Public wUnit As MyUnits
        Public A As Double
        Public B As Double
        Public C As Double
        Public D As Double
        Public E As Double
        Public F As Double

        Public Sub New(nUnit As MyUnits, nA As Double, nB As Double, nC As Double, nD As Double, nE As Double, nF As Double)
            wUnit = nUnit
            A = nA
            B = nB
            C = nC
            D = nD
            E = nE
            F = nF
        End Sub

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Affine")
            '
            res.AppendChild(wUnit.toXML(wDoc))
            res.SetAttribute("A", A)
            res.SetAttribute("B", B)
            res.SetAttribute("C", C)
            res.SetAttribute("D", D)
            res.SetAttribute("E", E)
            res.SetAttribute("F", F)
            '
            Return res
        End Function

    End Class

    Public Class MyBoundParam
        Public x1 As Double
        Public y1 As Double
        Public x2 As Double
        Public y2 As Double

        Public Sub New(nx1 As Double, ny1 As Double, nx2 As Double, ny2 As Double)
            x1 = nx1
            y1 = ny1
            x2 = nx2
            y2 = ny2
        End Sub

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Bounds")
            '
            res.SetAttribute("x1", x1)
            res.SetAttribute("y1", y1)
            res.SetAttribute("x2", x2)
            res.SetAttribute("y2", y2)
            '
            Return res
        End Function

    End Class

    Public Class MyEllipsoidDesc
        Public id As Integer
        Public Name As String
        Public a As Double
        Public f1 As Double
        Public EPSG As Integer
        Public proj As String

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Ellipsoid")
            '
            res.SetAttribute("id", id)
            res.SetAttribute("Name", Name)
            res.SetAttribute("a", a)
            res.SetAttribute("f1", f1)
            res.SetAttribute("EPSG", EPSG)
            '
            Return res
        End Function

        Public Function toProj() As String
            Dim res As String
            If proj = "" Then
                res = "+a=" & a & " +f=" & f1
            Else
                res = "+ellps=" & proj
            End If
            '
            Return res
        End Function

    End Class

    Public Class Datum3
        Public id As Integer
        Public Name As String
        Public Ellipsoid As MyEllipsoidDesc
        Public X As Double
        Public Y As Double
        Public Z As Double
        Public EPSG As Integer
        Public proj As String

        Public Sub New()
            '
        End Sub

        Public Sub New(nid As Integer, nel As Integer, nx As Double, ny As Double, nz As Double, EllipsoidList As List(Of MyEllipsoidDesc))
            id = nid
            Name = nid
            Ellipsoid = EllipsoidList.Find(Function(x) x.id = nel)
            X = nx
            Y = ny
            Z = nz
        End Sub

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Datum3")
            '
            res.AppendChild(Ellipsoid.toXML(wDoc))
            res.SetAttribute("id", id)
            res.SetAttribute("Name", Name)
            res.SetAttribute("X", X)
            res.SetAttribute("Y", X)
            res.SetAttribute("Z", Z)
            res.SetAttribute("EPSG ", EPSG)
            '
            Return res
        End Function

        Public Function toProj() As String
            Dim res As String
            If proj = "" Then
                res = Ellipsoid.toProj & " +towgs84=" & X & "," & Y & "," & Z
            Else
                res = "+datum=" & proj
            End If
            '
            Return res
        End Function

    End Class

    Public Class Datum8
        Inherits Datum3
        '
        Public RotationX As Double
        Public RotationY As Double
        Public RotationZ As Double
        Public Scale_ppm As Double
        Public Prime_Meridian As Double

        Public Sub New()
            '
        End Sub

        Public Sub New(nid As Integer, nel As Integer, nx As Double, ny As Double, nz As Double, rx As Double, ry As Double, rz As Double, nsp As Double, pm As Double, EllipsoidList As List(Of MyEllipsoidDesc))
            MyBase.New(nid, nel, nx, ny, nz, EllipsoidList)
            RotationX = rx
            RotationY = ry
            RotationZ = rz
            Scale_ppm = nsp
            Prime_Meridian = pm
        End Sub

        Public Overloads Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Datum8")
            '
            res.AppendChild(Ellipsoid.toXML(wDoc))
            res.SetAttribute("id", id)
            res.SetAttribute("Name", Name)
            res.SetAttribute("X", X)
            res.SetAttribute("Y", X)
            res.SetAttribute("Z", Z)
            res.SetAttribute("RotationX", RotationX)
            res.SetAttribute("RotationY", RotationY)
            res.SetAttribute("RotationZ", RotationZ)
            res.SetAttribute("Scale_ppm", Scale_ppm)
            res.SetAttribute("Prime_Meridian", Prime_Meridian)
            res.SetAttribute("EPSG", EPSG)
            '
            Return res
        End Function

        Public Overloads Function toProj() As String
            Dim res As String
            If proj = "" Then
                res = Ellipsoid.toProj & " +towgs84=" & X & "," & Y & "," & Z & "," & (-1 * RotationX) & "," & (-1 * RotationY) & "," & (-1 * RotationZ) & "," & Scale_ppm
                If Prime_Meridian <> 0 Then res = res & " +pm=" & Prime_Meridian
            Else
                res = "+datum=" & proj
            End If
            '
            Return res
        End Function

    End Class

    Public Class MyParamDesc
        Implements IComparable
        '
        Public id As Integer
        Public EPSG As Integer
        Public Desc As String
        Public Name As String
        Public Value As Object = Nothing
        Public WKT_Name As String
        Public proj_Name As String
        '
        Public Function CompareTo(obj As Object) As Integer Implements IComparable.CompareTo
            Dim wObj As MyParamDesc = obj
            If wObj.id > id Then
                Return -1
            ElseIf wObj.id = id Then
                Return 0
            Else
                Return 1
            End If
        End Function

        Public Function Clone() As MyParamDesc
            Dim res As New MyParamDesc
            res.id = id
            res.EPSG = EPSG
            res.Desc = Desc
            res.Name = Name
            res.Value = Value
            res.WKT_Name = WKT_Name
            res.proj_Name = proj_Name
            Return res
        End Function

        Public Overloads Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("ParamDesc")
            '
            'res.AppendChild(Ellipsoid.toXML(wDoc))
            res.SetAttribute("id", id)
            res.SetAttribute("Name", Name)
            res.SetAttribute("WKT_Name", WKT_Name)
            res.SetAttribute("proj_Name", proj_Name)
            res.SetAttribute("Desc", Desc)
            res.SetAttribute("EPSG", EPSG)
            Select Case id
                Case 1
                    Select Case Value.GetType.Name
                        Case "Datum3"
                            Dim wD As Datum3 = Value
                            res.AppendChild(wD.toXML(wDoc))
                        Case "Datum8"
                            Dim wD As Datum8 = Value
                            res.AppendChild(wD.toXML(wDoc))
                    End Select
                Case 2
                    Dim wU As MyUnits = Value
                    res.AppendChild(wU.toXML(wDoc))
                Case Else
                    res.SetAttribute("Value", Value)
            End Select
            '
            Return res
        End Function

        Public Function toProj() As String
            Dim res As String = ""
            Select Case id
                Case 1
                    Select Case Value.GetType.Name
                        Case "Datum3"
                            Dim wD As Datum3 = Value
                            res = wD.toProj
                        Case "Datum8"
                            Dim wD As Datum8 = Value
                            res = wD.toProj
                    End Select
                Case 2
                    Dim wU As MyUnits = Value
                    res = wU.toProj  '!!!!
                Case Else
                    res = proj_Name.Trim & "=" & Value
            End Select
            '
            Return res
        End Function

    End Class

    Public Class MyProjectionDesc
        Public id As Integer
        Public EPSG As Integer
        Public Name As String
        Public ruName As String
        Public WKT_Name As String
        Public proj_Name As String
        '
        Public ParamList As List(Of MyParamDesc)
        '
        Public Sub New()
            ParamList = New List(Of MyParamDesc) '!!!?
        End Sub

        Public Function Clone() As MyProjectionDesc
            Dim res As New MyProjectionDesc
            res.id = id
            res.EPSG = EPSG
            res.Name = Name
            res.ruName = ruName
            res.WKT_Name = WKT_Name
            res.proj_Name = proj_Name
            For Each pd In ParamList
                res.ParamList.Add(pd.Clone)
            Next
            Return res
        End Function

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Projection")
            '
            res.SetAttribute("id", id)
            res.SetAttribute("Name", Name)
            res.SetAttribute("ruName", ruName)
            res.SetAttribute("WKT_Name", WKT_Name)
            res.SetAttribute("proj_Name", proj_Name)
            res.SetAttribute("EPSG", EPSG)
            '
            For Each par In ParamList
                res.AppendChild(par.toXML(wDoc))
            Next
            '
            Return res
        End Function

        Public Function toProj() As String
            If proj_Name <> "" Then
                Dim res As String = "+proj=" & proj_Name '????!!!!
                '
                For Each pd In ParamList
                    res = res & " " & pd.toProj
                Next
                '
                Return res
            End If
            Return ""
        End Function

    End Class

    Public Class MyProjLib
        Public UnitList As List(Of MyUnits)
        Public EllipsoidList As List(Of MyEllipsoidDesc)
        Public Datum3List As List(Of Datum3)
        Public Datum8List As List(Of Datum8)
        Public ParamList As List(Of MyParamDesc)
        Public ProjectionList As List(Of MyProjectionDesc)

        Private Sub New()
            'Ellipsoid = New List(Of MyEllipsoidDesc)
        End Sub

        Public Shared Function Load_mdb(dbPath As String) As MyProjLib
            Dim wDL As New MyDataLink(dbPath)
            Dim res As New MyProjLib
            '
            wDL.conn.Open()
            res.UnitList = wDL.LoadUnitList()
            res.EllipsoidList = wDL.LoadEllipsoidList
            res.Datum3List = wDL.LoadDatum3List(res.EllipsoidList)
            res.Datum8List = wDL.LoadDatum8List(res.EllipsoidList)
            res.ParamList = wDL.LoadParamList()
            res.ProjectionList = wDL.LoadProjectionList()
            wDL.LoadProjParamLinks(res.ParamList, res.ProjectionList)
            wDL.conn.Close()
            '
            Return res
        End Function

        'MyProjectionDesc
        Public Function GetProjectionDesc(pid As Integer) As MyProjectionDesc
            Dim wPD As MyProjectionDesc = ProjectionList.Find(Function(x) x.id = pid) 'idProjectionTypes
            Return wPD.Clone
        End Function

    End Class

    Public Class MyCS
        Public Name As String
        Public Affin As Boolean = False
        Public wAffineParam As MyAffineParam = Nothing
        Public Bounds As Boolean = False
        Public wBoundParam As MyBoundParam = Nothing
        Public wPD As MyProjectionDesc
        Public wDatum As Object '???

        Public Shared Function ReadCSString(CSStr As String, wDL As MyDataLink) As MyCS
            If (CSStr.Length > 0) And (CSStr.Contains("""")) Then
                Return New MyCS(CSStr, wDL)
            Else
                Return Nothing
            End If
        End Function

        Public Shared Function ReadCSString(CSStr As String, wLib As MyProjLib) As MyCS
            If (CSStr.Length > 0) And (CSStr.Contains("""")) Then
                Return New MyCS(CSStr, wLib)
            Else
                Return Nothing
            End If
        End Function

        Public Sub New(CSStr As String)
            Dim NameEnd As Integer = CSStr.LastIndexOf("""")
            Name = CSStr.Substring(1, NameEnd - 1)
            'ParseParam(CSStr.Substring(NameEnd + 2, CSStr.Length - (NameEnd + 2)))
        End Sub

        Public Sub New(CSStr As String, wDL As MyDataLink)
            Dim NameEnd As Integer = CSStr.LastIndexOf("""")
            Name = CSStr.Substring(1, NameEnd - 1)
            ParseParam(CSStr.Substring(NameEnd + 2, CSStr.Length - (NameEnd + 2)), wDL)
        End Sub

        Public Sub New(CSStr As String, wLib As MyProjLib)
            Dim NameEnd As Integer = CSStr.LastIndexOf("""")
            Name = CSStr.Substring(1, NameEnd - 1)
            ParseParam(CSStr.Substring(NameEnd + 2, CSStr.Length - (NameEnd + 2)), wLib)
        End Sub

        Public Sub ParseParam(wStr As String, wDL As MyDataLink)
            Dim wStrArr() As String = wStr.Split(",")
            Dim idProjectionTypes As Integer = wStrArr(0)
            Select Case idProjectionTypes
                Case Is > 2999
                    Me.Affin = True
                    Me.Bounds = True
                    idProjectionTypes -= 3000
                Case Is > 1999
                    Me.Bounds = True
                    idProjectionTypes -= 2000
                Case Is > 999
                    Me.Affin = True
                    idProjectionTypes -= 1000
            End Select
            wPD = wDL.LoadProjectionDesc(idProjectionTypes)
        End Sub

        Public Sub ParseParam(wStr As String, wLib As MyProjLib)
            Dim wStrArr() As String = wStr.Split(",")
            Dim kp As Integer = 0
            Dim idProjectionTypes As Integer = wStrArr(kp)
            kp += 1
            Select Case idProjectionTypes
                Case Is > 2999
                    Me.Affin = True
                    Me.Bounds = True
                    idProjectionTypes -= 3000
                Case Is > 1999
                    Me.Bounds = True
                    idProjectionTypes -= 2000
                Case Is > 999
                    Me.Affin = True
                    idProjectionTypes -= 1000
            End Select
            wPD = wLib.GetProjectionDesc(idProjectionTypes) 'idProjectionTypes
            If wPD IsNot Nothing Then
                wPD.ParamList.Sort()
                Dim idDatum As Integer
                Dim uid As Integer
                For Each par In wPD.ParamList
                    Select Case par.id
                        Case 1
                            idDatum = wStrArr(kp)
                            Select Case idDatum
                                Case Is < 999
                                    wDatum = wLib.Datum3List.Find(Function(x) x.id = idDatum)
                                Case Is = 999
                                    wDatum = New Datum3(idDatum, wStrArr(kp + 1), wStrArr(kp + 2), wStrArr(kp + 3), wStrArr(kp + 4), wLib.EllipsoidList)
                                    kp += 4
                                Case Is < 9999
                                    wDatum = wLib.Datum8List.Find(Function(x) x.id = idDatum)
                                Case Is = 9999
                                    wDatum = New Datum8(idDatum, wStrArr(kp + 1), wStrArr(kp + 2), wStrArr(kp + 3), wStrArr(kp + 4), wStrArr(kp + 5), wStrArr(kp + 6), wStrArr(kp + 7), wStrArr(kp + 8), wStrArr(kp + 9), wLib.EllipsoidList)
                                    kp += 9
                            End Select
                            par.Value = wDatum
                        Case 2
                            uid = wStrArr(kp)
                            par.Value = wLib.UnitList.Find(Function(x) x.id = uid)
                        Case 11
                        Case Else
                            par.Value = Double.Parse(wStrArr(kp))
                    End Select
                    kp += 1
                Next
            End If
            '
            If Me.Affin Then
                Dim AU As Integer = wStrArr(kp)
                Dim nUnit As MyUnits = wLib.UnitList.Find(Function(u) u.id = AU)
                kp += 1
                Me.wAffineParam = New MyAffineParam(nUnit, wStrArr(kp), wStrArr(kp + 1), wStrArr(kp + 2), wStrArr(kp + 3), wStrArr(kp + 4), wStrArr(kp + 5))
                kp += 6
            End If
            If Me.Bounds Then
                Me.wBoundParam = New MyBoundParam(wStrArr(kp), wStrArr(kp + 1), wStrArr(kp + 2), wStrArr(kp + 3))
                kp += 4 '?
            End If
        End Sub

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("CS")
            res.SetAttribute("Name", Name)
            '!!!Param
            res.AppendChild(wPD.toXML(wDoc))
            If Affin Then res.AppendChild(wAffineParam.toXML(wDoc))
            If Bounds Then res.AppendChild(wBoundParam.toXML(wDoc))
            Return res
        End Function

        Public Function toProj() As String
            Dim res As String = wPD.toProj
            '
            Return res
        End Function

    End Class

    Public Class CSCategory
        Public CName As String
        Public CCSList As List(Of MyCS)

        Public Shared Function ReadCSCategory(wStream As StreamReader, wDL As MyDataLink) As CSCategory
            If Not wStream.EndOfStream Then
                Dim wStr As String = wStream.ReadLine()
                If (wStr <> "") Then
                    Dim res As New CSCategory(wStr)
                    Do While (wStr <> "") And (Not wStream.EndOfStream)
                        wStr = wStream.ReadLine()
                        wStr = wStr.Trim
                        res.AddCS(wStr, wDL)
                    Loop
                    Return res
                End If
            End If
            Return Nothing
        End Function

        Public Shared Function ReadCSCategory(wStream As StreamReader, wLib As MyProjLib) As CSCategory
            If Not wStream.EndOfStream Then
                Dim wStr As String = wStream.ReadLine()
                If (wStr <> "") Then
                    Dim res As New CSCategory(wStr)
                    Do While (wStr <> "") And (Not wStream.EndOfStream)
                        wStr = wStream.ReadLine()
                        wStr = wStr.Trim
                        res.AddCS(wStr, wLib)
                    Loop
                    Return res
                End If
            End If
            Return Nothing
        End Function

        Public Sub New(NName As String)
            CName = NName
            CCSList = New List(Of MyCS)
        End Sub

        Public Function AddCS(wStr As String, wDL As MyDataLink) As Boolean
            Dim nCS As MyCS = MyCS.ReadCSString(wStr, wDL)
            If nCS IsNot Nothing Then
                CCSList.Add(nCS)
                Return True
            Else
                Return False
            End If
        End Function

        Public Function AddCS(wStr As String, wLib As MyProjLib) As Boolean
            Dim nCS As MyCS = MyCS.ReadCSString(wStr, wLib)
            If nCS IsNot Nothing Then
                CCSList.Add(nCS)
                Return True
            Else
                Return False
            End If
        End Function

        Public Function toXML(wDoc As XmlDocument) As XmlElement
            Dim res As XmlElement = wDoc.CreateElement("Category")
            res.SetAttribute("Name", CName)
            For Each cs In Me.CCSList
                res.AppendChild(cs.toxml(wDoc))
            Next
            Return res
        End Function

        Public Function toProj() As String
            Dim wSB As New StringBuilder
            '
            wSB.AppendLine(CName)
            For Each cs In CCSList
                wSB.AppendLine(cs.Name)
                wSB.AppendLine(cs.toProj)
            Next
            wSB.AppendLine("")
            '
            Return wSB.ToString
        End Function

    End Class

    Public Class PrjFile
        Public CategoriesList As List(Of CSCategory)
        Public wDL As MyDataLink
        Public wLib As MyProjLib        '= MyProjLib.Load_mdb("F:\progi\VS2012\mprj1\mprj1\bin\Debug\MapInfo.mdb")

        Protected Friend Sub New()
            CategoriesList = New List(Of CSCategory)
        End Sub

        Public Shared Function ReadFile(fName As String) As PrjFile
            'Try
            Dim wStream As New StreamReader(fName, System.Text.Encoding.Default)
            Dim res As New PrjFile
            Dim wCt As CSCategory
            res.wLib = MyProjLib.Load_mdb("F:\progi\VS2012\mprj1\mprj1\bin\Debug\MapInfo.mdb")
            Do While (Not wStream.EndOfStream)
                wCt = CSCategory.ReadCSCategory(wStream, res.wLib)
                If wCt IsNot Nothing Then res.CategoriesList.Add(wCt)
            Loop
            Return res
        End Function

        Public Shared Function ReadFileDL(fName As String) As PrjFile
            'Try
            Dim wStream As New StreamReader(fName, System.Text.Encoding.Default)
            Dim res As New PrjFile
            Dim wCt As CSCategory
            res.wDL = MyDataLink.Open("F:\progi\VS2012\mprj1\mprj1\bin\Debug\MapInfo.mdb")
            Do While (Not wStream.EndOfStream)
                wCt = CSCategory.ReadCSCategory(wStream, res.wDL)
                If wCt IsNot Nothing Then res.CategoriesList.Add(wCt)
            Loop
            Return res
        End Function

        Public Sub SaveToXML(fName As String)
            Dim wDoc As New XmlDocument
            wDoc.LoadXml("<MyPrjFile/>")
            For Each csc In CategoriesList
                wDoc.DocumentElement.AppendChild(csc.toxml(wDoc))
            Next
            wDoc.Save(fName)
        End Sub

        Public Sub SaveToProj(fName As String)
            Dim wSB As New StringBuilder
            For Each csc In CategoriesList
                wSB.Append(csc.toProj)
            Next
            'wSB .
            Using outfile As New StreamWriter(fName)
                outfile.Write(wSB.ToString())
            End Using
        End Sub

    End Class

End Module
